/*
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 * Amiga codec
 *
 * Copyright (C) 1998-2002 Michael Krause
 * 
 * >> modified by Andreas Hein (a_hein@gmx.net):
 * >> added support for 984k/1968k Disks (diskspare.device)  
 * >> -read only at this time
 * >> native Amiga-code  by Klaus Deppisch.
 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <asm/io.h>

#include "cwfloppy.h"
#include "encoder-common.h"

#if LINUX_VERSION_CODE >= 0x020400
#include <linux/slab.h>
#else
#include <linux/malloc.h>
#endif
struct amiga_header
{
  unsigned char magic;
  unsigned char track;
  unsigned char sect;
  unsigned char ord;
  unsigned char labels[16];
  unsigned long hdrchk;
  unsigned long datachk;
};

struct diskspare_header
{
  unsigned short gap2;
  unsigned char track;
  unsigned char sect;
  unsigned short chk;
};

static int encode_tmp[2];
static int mfmtmp[2];
static void (*putfunc) (const char *, int, int);



unsigned char
mfm_sub_decode (unsigned short word)
{
  int i;
  register unsigned char byte = 0;

  for (i = 0; i < 16; i += 2)
    byte |= (((word & (1 << i)) > 0) << (i / 2));

  return byte;
}

static int
amiga_decode_mfm (int iobase, int numbytes, void *buf, int modus,
		  unsigned short *csum)
{
  char b;
  int c = 0;
  unsigned char *sbuf = buf;
  unsigned short word;

  int tt1 = mfmtmp[0], tt2 = mfmtmp[1];

  while (c < numbytes && (b = inb (iobase)) >= 0)
    {
      b = threshtab[(int) b];
      tt1 = (tt1 << b) + 1;	/* collect bits from device */
      tt2 += b;

      if (tt2 >= 16)
	{
	  tt2 -= 16;
	  word = (tt1 >> tt2);

	  /* checksum calculation */
	  if (modus == 0x81)	/* ds-checksum -init */
	    {
	      modus = 1;
	      *csum ^= word & 0x7fff;
	    }
	  else if (csum != NULL)
	    *csum ^= word;

	  /* mfm-decoding */
	  if (modus == 1)	/* modus=1: decode word by word 
				   even/odd -orig.bits separated  */
	    {
	      word &= 0x5555;
	      if (c < numbytes / 2)
		{
		  *sbuf++ = (word >> 8) << 1;
		  *sbuf++ = (word & 0xff) << 1;
		}
	      else
		{
		  *sbuf++ |= (word >> 8);
		  *sbuf++ |= (word & 0xff);
		}
	    }

	  else
	    /* modus=0 : decode word to byte , native mfm-decoding  */
	    *sbuf++ = mfm_sub_decode (word);

	  c++;
	  if (modus == 1 && (c == numbytes / 2))
	    sbuf = buf;
	}
    }

  mfmtmp[0] = tt1;
  mfmtmp[1] = tt2;
  return c == numbytes;

}

/* Who needs a real CRC anyway? :)*/

static unsigned long
checksum (unsigned long *addr, int len)
{
  unsigned long csum = 0;
  len /= sizeof (*addr);
  while (len-- > 0)
    csum ^= *addr++;
  csum = ((csum >> 1) & 0x55555555) ^ (csum & 0x55555555);
  return csum;
}

static int
find_sync (int iobase)
{
  static const char sync[] = {
    2 + 2, 1 + 2, 2 + 2, 1 + 2, 0 + 2,
    2 + 2, 1 + 2, 2 + 2, 1 + 2, -1
  };
  const char *synccmp = sync;
  char b;
  while ((b = inb (iobase)) >= 0)
    {
      if (threshtab[(int) b] != *synccmp)
	{
	  synccmp = sync;
	}
      else
	{
	  synccmp++;
	  if (*synccmp == -1)
	    return 1;
	}
    }

  return 0;
}

unsigned short
byteswap_w (unsigned short word)
{
  word = (word << 8) | (word >> 8);	/* byteswap-corr. */
  return word;
}



static int
amiga_decode (struct cw_drive_struct *d, int side, int try)
{
  int drivetype = d->d->type;
  static unsigned char amiga_thresholds[] = { 0x22, 0x30 };	/* default 3.5"-Drive */
  const int iobase = d->c->c.iobase;
  const unsigned int maxsecs = cw_formats[d->format].max_secs_per_track;
  int todo, i = 0;
  struct amiga_header hdr;
  struct diskspare_header ds_hdr;
  unsigned char *dest;
  unsigned long crc;
  unsigned short mfmcsum;

  if (drivetype == 1)
    {
      amiga_thresholds[0] = 0x22;
      amiga_thresholds[1] = 0x30;
    }				// 3.5"-Drive 
  else
    {
      amiga_thresholds[0] = 0x1b;
      amiga_thresholds[1] = 0x26;
    }				//  5.25"-Drive


  todo = (1 << maxsecs) - 1;	/* bit mask of decoded sectors */
  codec_init_threshtab (try, amiga_thresholds);
  mfmcsum = 0;

repeat:			/* oh, this is sooo evil. */

  if (!find_sync (iobase))
    goto out;
  mfmtmp[0] = mfmtmp[1] = 0;
  switch (maxsecs)
    {
    case 10:			/// Amiga: trackdisk.device
    case 11:
    case 20:
    case 22:

      if (!amiga_decode_mfm (iobase, 4, &hdr.magic, 1, NULL))
	goto out;
      if (!amiga_decode_mfm (iobase, 16, &hdr.labels, 1, NULL))
	goto out;
      if (!amiga_decode_mfm (iobase, 4, &hdr.hdrchk, 1, NULL))
	goto out;
      if (!amiga_decode_mfm (iobase, 4, &hdr.datachk, 1, NULL))
	goto out;
      crc = checksum ((unsigned long *) &hdr, 20);
      if (hdr.hdrchk != crc)
	{
	  cw_codec_error (1, "header checksum mismatch %08x:%08x",
			  hdr.hdrchk, crc);
	  return 0;
	}
      if (hdr.magic != 0xff)
	{
	  cw_codec_error (2, "hdr.magic != 0xff");
	  return 0;
	}

      if (hdr.track != 2 * d->d->track + side)
	{
	  cw_codec_error (2, "hdr.track mismatch (%d)", hdr.track);
	  return 0;
	}
      break;

    case 12:			/// Amiga: diskspare.device
    case 24:
      /* gap2 (0x00) between sync and sectorinfo  */
      if (!amiga_decode_mfm (iobase, 1, &ds_hdr.gap2, 0, NULL))
	goto out;
      if (ds_hdr.gap2)
	{
	  cw_codec_error (2, "gap2 wrong (%d) not 0x00", ds_hdr.gap2);
	  return 0;
	}

      /* sectorinfo: track.b, sector.b, checksum.w */
      if (!amiga_decode_mfm (iobase, 4, &ds_hdr.track, 1, NULL))
	goto out;
      if (ds_hdr.track != 2 * d->d->track + side)
	{
	  cw_codec_error (2,
			  "hdr.track mismatch: found track=%d, sector=%d, checksum=%d",
			  ds_hdr.track, ds_hdr.sect, ds_hdr.chk);
	  return 0;
	}

      hdr.track = ds_hdr.track;
      hdr.sect = ds_hdr.sect;
      break;
    }

  if (hdr.sect >= maxsecs)
    {
      cw_codec_error (2, "hdr.sect (%d) out of range (0-%d), track=%d\n",
		      hdr.sect, maxsecs - 1, hdr.track);
      return 0;
    }

  if (!(todo & (1 << hdr.sect)))
    goto repeat;		/* we already have this sector */

  dest = d->trackbuffer + hdr.sect * 512;	/* dest.- data buffer */

  switch (maxsecs)
    {
    case 12:
    case 24:

      /* 512 bytes data diskspare */
      mfmcsum = 0;
      /* the first 4 bytes data , mfm-csum init */
      if (!amiga_decode_mfm (iobase, 4, dest, 0x81, &mfmcsum))
	goto out;

      /*  508 bytes data */
      for (i = 1; i < 128; i++)
	if (!amiga_decode_mfm (iobase, 4, dest + 4 * i, 1, &mfmcsum))
	  goto out;
      mfmcsum = byteswap_w (mfmcsum);	/* byteswap-corr. */
      if (ds_hdr.chk != mfmcsum)
	{
	  cw_codec_error (1,
			  "ds_data checksum mismatch hdrinfo=%04x:calc=%04x",
			  ds_hdr.chk, mfmcsum);
	  return 0;
	}
      break;

    case 10:
    case 11:
    case 20:
    case 22:

      /* 512 bytes data trackdisk.dev. */
      if (!amiga_decode_mfm (iobase, 512, dest, 1, NULL))
	goto out;
      crc = checksum ((unsigned long *) dest, 512);
      if (hdr.datachk != crc)
	{
	  cw_codec_error (1, "data checksum mismatch, hdrinfo=%08x:calc=%08x",
			  hdr.datachk, crc);
	  return 0;
	}

      break;

    default:
      cw_codec_error (0,
		      "number of sectors (%d) wrong , what happened \?\?\?\?!",
		      maxsecs);
      return 0;
    }


  todo &= ~(1 << hdr.sect);	/*  remove sector found from todo-list */
  if (!todo)			/*  are all sectors found ? then fisnish */
    return 1;
  goto repeat;
out:
  cw_codec_error (0, todo == (1 << maxsecs) - 1 ?
		  "can't find sync" :
		  "track buffer incomplete sector=%d todo=%8x", hdr.sect,
		  todo);
  return 0;
}

static void
do_mfm_csum (int b, unsigned short *mfmcsum)
{
  int tt1, tt2;
  unsigned short word;
  tt1 = mfmtmp[0];
  tt2 = mfmtmp[1];

  tt1 = (tt1 << 2) | b;		/* collect bits */
  tt2 += 2;


  if (tt2 == 16)
    {
      word = (unsigned short) tt1;
      tt1 = tt2 = 0;
      // printk("word=0x%4x\n",word);

      /* checksum calculation */
      *mfmcsum ^= word;
    }
  mfmtmp[0] = tt1;
  mfmtmp[1] = tt2;
}

static void
diskspare_mfm_checksum (const unsigned char *data, int num, int shift,
			unsigned short *mfmcsum)
{
  int prevdatabit, nextdatabit = encode_tmp[0];
  unsigned char x = 0;

  num *= 4;

  while (num)
    {
      if (!(num-- & 3))
	x = (*data++) << shift;
      else
	x <<= 2;

      prevdatabit = nextdatabit;
      nextdatabit = x >> 7;

      if (nextdatabit)		/* next data bit is 1 -> write 01 */
	do_mfm_csum (1, mfmcsum);

      else			/* next data bit is 0 */
	{
	  if (prevdatabit)
	    do_mfm_csum (0, mfmcsum);	/* write 00 */
	  else
	    do_mfm_csum (2, mfmcsum);	/* write 10 */
	}
    }
  encode_tmp[0] = nextdatabit;
}



/* The Amiga does MFM encoding for the even bits and odd bits
   separately. This has the advantage that the en-/decoding (which, on
   the Amiga hardware, is not done by the floppy controller itself)
   can be done efficiently using the CPU (or the blitter) without
   having to shuffle bits around too much.

   This routine does the encoding without any intermediate buffer. It
   may well be possible to find a faster algorithm. */


static void
amiga_encode_mfm (int iobase,
		  const unsigned char
		  *data, int num, const unsigned char *mfmenc, int shift)
{
  int prevdatabit, nextdatabit = encode_tmp[0], nullcnt = encode_tmp[1];
  unsigned char x = 0;
  num *= 4;
  while (num)
    {
      if (!(num-- & 3))
	x = (*data++) << shift;
      else
	x <<= 2;
      prevdatabit = nextdatabit;
      nextdatabit = x >> 7;
      if (prevdatabit == 0)
	{
	  /* previous data bit was 0, nullcnt is thus 1 or 2 */
	  if (nextdatabit == 0)
	    {
	      /* next data bit is 0 -> clock bit is 1 -> write 01 or 001 and
	         remember one null. */
	      PUT (nullcnt - 1);
	      nullcnt = 1;
	    }
	  else
	    {
	      /* next data bit is 1 -> clock bit is 0 -> write 001 or 0001
	         and clear the null counter. */
	      PUT (nullcnt);
	      nullcnt = 0;
	    }
	}
      else
	{
	  /* previous data bit was 1, nullcnt is thus 0. */
	  if (nextdatabit == 0)
	    {
	      /* next data bit 0 -> clock bit also 0 -> 2 nulls in a row */
	      nullcnt = 2;
	    }
	  else
	    {
	      /* next data bit 1 -> clock bit is 0 -> write "01" and
	         leave nullcnt = 0. */
	      PUT (0);
	    }
	}
    }

  encode_tmp[0] = nextdatabit;
  encode_tmp[1] = nullcnt;
}

static inline void
amiga_encode_sector (int iobase, int track, int sec, int maxsecs,
		     unsigned char *data, const unsigned char *mfmenc)
{
  static const unsigned char mark1[] = { 0, 0, 0, 0 };
  int i;
  unsigned short mfmcsum;
  struct amiga_header hdr;
  struct diskspare_header ds_hdr;

  /* intro mark */
  amiga_encode_mfm (iobase, mark1, 4, mfmenc, 0);	/* gap1, 2 bytes */

  /* write two syncs 0x4489 */
  PUT (1);
  PUT (2);
  PUT (1);
  PUT (2);
  PUT (1);
  PUT (0);
  PUT (2);
  PUT (1);
  PUT (2);
  PUT (1);

  encode_tmp[1] = 0;
  encode_tmp[0] = 1;

  switch (maxsecs)
    {
    case 10:			/// Amiga:trackdisk.device
    case 11:
    case 20:
    case 22:
      hdr.magic = 0xFF;
      hdr.track = track;
      hdr.sect = sec;
      hdr.ord = maxsecs - sec;
      for (i = 0; i < 16; i++)
	hdr.labels[i] = 0;
      hdr.hdrchk =
	checksum ((unsigned long *) &hdr,
		  (char *) &hdr.hdrchk - (char *) &hdr);
      hdr.datachk = checksum ((unsigned long *) data, 512);
      amiga_encode_mfm (iobase, &hdr.magic, 4, mfmenc, 0);
      amiga_encode_mfm (iobase, &hdr.magic, 4, mfmenc, 1);
      amiga_encode_mfm (iobase, (unsigned char *) &hdr.labels, 16, mfmenc, 0);
      amiga_encode_mfm (iobase, (unsigned char *) &hdr.labels, 16, mfmenc, 1);
      amiga_encode_mfm (iobase, (unsigned char *) &hdr.hdrchk, 4, mfmenc, 0);
      amiga_encode_mfm (iobase, (unsigned char *) &hdr.hdrchk, 4, mfmenc, 1);
      amiga_encode_mfm (iobase, (unsigned char *) &hdr.datachk, 4, mfmenc, 0);
      amiga_encode_mfm (iobase, (unsigned char *) &hdr.datachk, 4, mfmenc, 1);
      amiga_encode_mfm (iobase, data, 512, mfmenc, 0);
      amiga_encode_mfm (iobase, data, 512, mfmenc, 1);
      break;

    case 12:			/// Amiga: diskspare.device
    case 24:

      ds_hdr.track = track;
      ds_hdr.sect = sec;

      /* diskspare is calculating its checksum on mfm-data */
      mfmcsum = 0x0000;
      mfmtmp[0] = 0;		// tt1=bitstring
      mfmtmp[1] = 0;		// tt2=bitcounter
      encode_tmp[0] = 1;	// prevdatabit

      for (i = 0; i < 128; i++)	/* encode long by long */
	{
	  diskspare_mfm_checksum (data + 4 * i, 4, 0, &mfmcsum);	/* calc.the checksum */
	  diskspare_mfm_checksum (data + 4 * i, 4, 1, &mfmcsum);
	}
	
      ds_hdr.chk = byteswap_w (mfmcsum);	/* byteswap-corr. */

      /* encoding data */
      encode_tmp[0] = 1;
      encode_tmp[1] = 0;
      amiga_encode_mfm (iobase, mark1, 2, mfmenc, 0);	/* gap2, 1 byte */
      amiga_encode_mfm (iobase, &ds_hdr.track, 4, mfmenc, 0);	/* write track,sector,chksum */
      amiga_encode_mfm (iobase, &ds_hdr.track, 4, mfmenc, 1);

      for (i = 0; i < 128; i++)	/* encode long by long */
	{
	  amiga_encode_mfm (iobase, data + 4 * i, 4, mfmenc, 0);
	  amiga_encode_mfm (iobase, data + 4 * i, 4, mfmenc, 1);
	}
      break;
    default:

    }
}

static void
amiga_encode (struct cw_drive_struct *d, int side)
{
  const int iobase = d->c->c.iobase;
  const unsigned int maxsecs = cw_formats[d->format].max_secs_per_track;
  int drivetype = d->d->type;
  int i;
  unsigned char *data = d->trackbuffer;
  static const unsigned char enddata[] = { 0, 0 };
  static unsigned char mfmenc[3];

  /* basic putfunc + pulselengths:  3.5"-Disk */
  putfunc = (d->d->track >= 40 && (maxsecs >= 20)) ? put_precomp : put_normal;
  mfmenc[0] = 0x80 - 0x1b;
  mfmenc[1] = 0x80 - 0x2a;
  mfmenc[2] = 0x80 - 0x38;


  /* adapt mfm-pulselenghts: increasing values -> shortening mfm-pulses on disk */
  if (drivetype == 1)		/* 3.5"-Disk */
    {
      if (maxsecs == 12)
	{
	  mfmenc[0] += 1;
	  mfmenc[1] += 1;
	  mfmenc[2] += 1;
	}

      if (maxsecs == 24)
	{
	  mfmenc[0] += 1;
	  mfmenc[1] += 1;
	  mfmenc[2] += 1;
	}
    }

  if (drivetype == 2)		/* 5.25"-Disk */
    {
      mfmenc[0] += 6;
      mfmenc[1] += 8;
      mfmenc[2] += 10;

      /* if (maxsecs == 12)	// Test-Area :-)
	{
	  mfmenc[0] += 1;
	  mfmenc[1] += 0;
	  mfmenc[2] += 0;
	}

      if (maxsecs == 24)
	{
	  mfmenc[0] += 1;
	  mfmenc[1] += 0;
	  mfmenc[2] += 0;
	} */
    }

  inb (iobase + 1);		/***** start writing track into Catweasel's memory *****/

  /* track-gap, 
     will be partly overwritten, according to the real physical track length which depends on the userdata
   */
  switch (maxsecs)
    {
    case 10:
    case 20:
      for (i = 0; i < 12000 * (1 + (maxsecs > 10)) * ((drivetype == 2) ? 1.2:1); i++)
	PUT (0);
      break;
    case 11:
    case 22:
      for (i = 0; i <  8000 * (1 + (maxsecs > 11)) * ((drivetype == 2) ? 1.2:1); i++)
	PUT (0);
      break;
    case 12:
    case 24:
      for (i = 0; i <  2000 * (1 + (maxsecs > 12)) * ((drivetype == 2) ? 1.2:1); i++)
	PUT (0);
    }

  encode_tmp[0] = 0;
  encode_tmp[1] = 1;

  /* the sectors */
  for (i = 0; i < maxsecs; i++, data += 512)
    amiga_encode_sector (iobase, 2 * d->d->track + side, i, maxsecs, data,
			 mfmenc);

  /* end mark */
  amiga_encode_mfm (iobase, enddata, 2, mfmenc, 0);

  outb (128, iobase + 0);	/****** stop writing track into Catweasel's memory ******/


#if ENCODE_CALC_WRITETIME
  calc_time (iobase);
#endif
}

struct cw_codec cw_codec_amiga = {
  amiga_decode, amiga_encode
};
